#include <cave_ogl.h>
#include <GL/glu.h>	// Header File For The GLu32 Library
#include <stdio.h>	  // Header file for standard file i/o.
#include <stdlib.h>
#include <iostream>
#include "jpeglib.h"

char *buffer=0;
int size=256*256*3; // Texture 256x256, RGB
void frameFunc();
void DrawGLScene();
void InitGL();
void * JPEGImageLoad(char *filename, char *image);
int *texture=0;
bool changed=0;
//CAVELOCK lock;
float xrot, yrot, zrot; 
void tangle(char *buffer,int size,int strength);
int tangle_strength=10;
int main(int argc, char **argv) 
{  
	CAVEConfigure(&argc,argv,NULL);
	if (argc>1) tangle_strength=atoi(argv[1]);
	std::cout<<"Allocating buffer ... ";
	buffer=(char*)CAVEAllocDisplayData(sizeof(char)*size);
	if (!buffer) {
		std::cout<<"Failed"<<std::endl;
		CAVEExit();
		return 0;
	}
	std::cout<<"OK"<<std::endl;
	if (CAVEDistribMaster()) { // Only Master should read texture file
		std::cout<<"Loading texture...";
		if (!JPEGImageLoad("uwajima.jpg",(char*)buffer)) {
			std::cout<<"Failed"<<std::endl;
			CAVEExit();
			return 0;
		}
		std::cout<<"OK"<<std::endl;
	}
	texture=(int*)CAVEMalloc(sizeof(int));
	std::cout<<"Setting up CAVE callbacks"<<std::endl;
	CAVEInitApplication((CAVECALLBACK)InitGL,0);
	CAVEDisplay((CAVECALLBACK)DrawGLScene,0);
	CAVEFrameFunction((CAVECALLBACK)frameFunc,0);
	
	
	
	std::cout << "Opening DistribConnection " << std::endl;
	CAVEDistribOpenConnection(1);
	CAVEDistribOpenConnection(2);

	CAVEDistribBarrier(1);
	if (CAVEDistribMaster()) {
		std::cout<<"Sending msg to slaves"<<std::endl;
		CAVEDistribWrite(2,(void*)"Hellu!",7);
		std::cout<<"Msg sent."<<std::endl;
    } else {
		char str[100];
		std::cout<<"Receiving msg"<<std::endl;
		CAVEDistribRead(2,(void*)str,7);
		std::cout<<"Msg: "<<str<<std::endl;
    }
	std::cout << "Finished" << std::endl;
    bool data_passed=0;
    CAVEDistribBarrier(1);
	
	CAVEInit();
	if (CAVEDistribMaster()) { // Only Master should pass display data
		std::cout << "Passing Display data" << std::endl;
		CAVEPassDisplayData(buffer,0);
	}
	std::cout << "Starting up main loop" << std::endl;	
 	if (CAVEDistribMaster()) {
		while (!CAVEgetbutton(CAVE_ESCKEY)) {
			CAVEUSleep(10);
			tangle(buffer,size,tangle_strength);
			CAVEPassDisplayData(buffer,0);
		}
	} else {
		while (!CAVESync->Quit) CAVEUSleep(15);
	}
	std::cout<< "Cleaning up." << std::endl;
	CAVEFree(texture);
	CAVEDistribCloseConnection(1);
	CAVEDistribCloseConnection(2);
	CAVEExit();
	return 0;
}



void frameFunc()
{
  
  if (CAVEMasterDisplay()) {
	static float prevtime = 0;
	float t = CAVEGetTime();
	float dt = t - prevtime;
	prevtime=t;
	float jx = CAVE_JOYSTICK_X;
	float jy = CAVE_JOYSTICK_Y;
	float _speed_ = 5.0f;
	if (fabs(jy)>0.2) {
	float wand[3];
	CAVEGetVector(CAVE_WAND_FRONT,wand);
	CAVENavTranslate(wand[0]*jy*dt*_speed_,
			wand[1]*jy*dt*_speed_,
			wand[2]*jy*dt*_speed_);
	}
	if (fabs(jx) > 0.2) {
	CAVENavRot(-jx*90*dt,'y');
	}
	CAVEDisplayBarrier();
  }
}

// Loads 256x256 texture from jpeg file. return buffer if succeeded, 0 otherwise
void *JPEGImageLoad(char *filename, char *image) {
  
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  FILE *file;		/* source file */
  if ((file = fopen(filename, "rb"))==NULL)
  {
	std::cout << "File Not Found" << filename << std::endl;
	return 0;
  }
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, file);
  (void) jpeg_read_header(&cinfo, TRUE);
  if (cinfo.image_width!=256 || cinfo.image_height != 256) {
	std::cout<<"Image " << filename<< " is not 256x256 jpeg file" << std::endl;
  }
  (void) jpeg_start_decompress(&cinfo);  
  char *buffer=(char*)malloc(sizeof(char)*256*3);
  while (cinfo.output_scanline < cinfo.output_height) {
	(void) jpeg_read_scanlines(&cinfo, (JSAMPLE**)&buffer, 1);
	(void*)memcpy((void*)(image+(256*3*cinfo.output_scanline)),(void*)buffer,256*3);
  }
  (void) jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  fclose(file);
  return image;
}

// Creates dummy texture
void LoadGLTextures() {	
	char *image=new char[size];
	
	memset(image,0x8F,size);
	glColor4f(1.0f,1.0f,1.0f,0.5f);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);	
	// Create Texture	
	glGenTextures(1, (GLuint*)&texture[0]);
	glBindTexture(GL_TEXTURE_2D, texture[0]);   // 2d texture (x and y size)

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); // scale linearly when image bigger than texture
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); // scale linearly when image smalled than texture
	std::cout << "Creating texture" << std::endl;
	// 2d texture, level of detail 0 (normal), 3 components (red, green, blue), x size from image, y size from image, 
	// border 0 (normal), rgb color data, unsigned byte data, and finally the data itself.
	glTexImage2D(GL_TEXTURE_2D, 0, 3, 256, 256, 0, GL_RGB, GL_UNSIGNED_BYTE, image);
	delete [] image;	
};


/* A general OpenGL initialization function.  Sets all of the initial parameters. */
//void InitGL(int Width, int Height)			// We call this right after our OpenGL window is created.
void InitGL()
{
	std::cout << "OpenGL initialization starting" << std::endl;
	LoadGLTextures();				// Load The Texture(s) 
	std::cout << "Textures loaded" << std::endl;
	glEnable(GL_TEXTURE_2D);			// Enable Texture Mapping
	glClearColor(.0f, .0f, 0.5f, 0.0f);		// Clear The Background Color To Blue 
	glClearDepth(1.0);				// Enables Clearing Of The Depth Buffer
	glDepthFunc(GL_LESS);			// The Type Of Depth Test To Do
	glEnable(GL_DEPTH_TEST);			// Enables Depth Testing
	glShadeModel(GL_SMOOTH);			// Enables Smooth Color Shading
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	std::cout << "OpenGL initialization finished" << std::endl;
}


/* The main drawing function. */
void DrawGLScene()
{
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 256, 256, GL_RGB, GL_UNSIGNED_BYTE,(GLvoid*) CAVEGetDisplayData(buffer,NULL));
//	std::cout << "\rFPS: " << *CAVEFramesPerSecond;	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Clear The Screen And The Depth Buffer

	CAVENavTransform();
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);	

	glTranslatef(0.0f,0.0f,-5.0f);			  // move 5 units into the screen.
	
	glRotatef(xrot,1.0f,0.0f,0.0f);		// Rotate On The X Axis
	glRotatef(yrot,0.0f,1.0f,0.0f);		// Rotate On The Y Axis
	glRotatef(zrot,0.0f,0.0f,1.0f);		// Rotate On The Z Axis

	glBindTexture(GL_TEXTURE_2D, texture[0]);   // choose the texture to use.

	glBegin(GL_QUADS);						// begin drawing a cube
	
	// Front Face (note that the texture's corners have to match the quad's corners)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);	// Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);	// Top Left Of The Texture and Quad
	
	// Back Face
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);	// Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);	// Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);	// Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);	// Bottom Left Of The Texture and Quad
	
	// Top Face
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);	// Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);	// Top Right Of The Texture and Quad
	
	// Bottom Face	   
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f, -1.0f, -1.0f);	// Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f, -1.0f, -1.0f);	// Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
	
	// Right face
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);	// Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);	// Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);	// Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
	
	// Left Face
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);	// Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);	// Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);	// Top Left Of The Texture and Quad
	
	glEnd();									// done with the polygon.

	xrot+=0.15f;						// X Axis Rotation	
	yrot+=0.15f;						// Y Axis Rotation
	zrot+=0.15f;						// Z Axis Rotation

}
void swap_char (char *a, char *b)
{
	char c=*a;
	*a=*b;
	*b=c;
}
void tangle(char *buffer,int size,int strength){
	for (int i=0;i<strength;i++) {
		swap_char (buffer+(int)((float)rand()*size/RAND_MAX),buffer+(int)((float)rand()*size/RAND_MAX));
	}
}
